@@ -1,24 +1,20 @@
 /*
- * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2010 Trustwave Holdings, Inc. (http://www.trustwave.com/)
- *
- * This product is released under the terms of the General Public Licence,
- * version 2 (GPLv2). Please refer to the file LICENSE (included with this
- * distribution) which contains the complete text of the licence.
- *
- * There are special exceptions to the terms and conditions of the GPL
- * as it is applied to this software. View the full text of the exception in
- * file MODSECURITY_LICENSING_EXCEPTION in the directory of this software
- * distribution.
- *
- * If any of the files related to licensing are missing or if you have any
- * other questions related to licensing please contact Trustwave Holdings, Inc.
- * directly using the email address support@trustwave.com.
- *
- */
+* ModSecurity for Apache 2.x, http://www.modsecurity.org/
+* Copyright (c) 2004-2011 Trustwave Holdings, Inc. (http://www.trustwave.com/)
+*
+* You may not use this file except in compliance with
+* the License.  You may obtain a copy of the License at
+*
+*     http://www.apache.org/licenses/LICENSE-2.0
+*
+* If any of the files related to licensing are missing or if you have any
+* other questions related to licensing please contact Trustwave Holdings, Inc.
+* directly using the email address security@modsecurity.org.
+*/
+
+#include "modsecurity.h"
 #include <sys/stat.h>
 
-#include "mod_security2_config.h"
 #include "re.h"
 #include "msc_logging.h"
 #include "httpd.h"
@@ -26,6 +22,8 @@
 #include "apr_global_mutex.h"
 #include "msc_util.h"
 
+#include "apr_version.h"
+#include <libxml/xmlversion.h>
 
 /**
  * Write the supplied data to the audit log (if the FD is ready), update
@@ -70,51 +68,6 @@ static int sec_auditlog_write(modsec_rec *msr, const char *data, unsigned int le
     return 1;
 }
 
-/**
- * Construct a log line in the vcombinedus format (see below).
- */
-char *construct_log_vcombinedus(modsec_rec *msr) {
-    const char *local_user, *remote_user;
-    const char *referer, *user_agent, *uniqueid;
-    const char *sessionid;
-
-    /* remote log name */
-    if (msr->remote_user == NULL) remote_user = "-";
-    else remote_user = msr->remote_user;
-
-    /* authenticated user */
-    if (msr->local_user == NULL) local_user = "-";
-    else local_user = msr->local_user;
-
-    /* unique id */
-    uniqueid = msr->txid;
-    if (uniqueid == NULL) uniqueid = "-";
-
-    /* referer */
-    referer = "-";
-    /* Logging Referer is a waste of space.
-    referer = (char *)apr_table_get(msr->request_headers, "Referer");
-    if (referer == NULL) referer = "-";
-    */
-
-    /* user agent */
-    user_agent = "-";
-    /* Logging User-Agent is a waste of space too.
-    user_agent = (char *)apr_table_get(msr->request_headers, "User-Agent");
-    if (user_agent == NULL) user_agent = "-";
-    */
-
-    /* sessionid */
-    sessionid = (msr->sessionid == NULL ? "-" : msr->sessionid);
-
-    return apr_psprintf(msr->mp, "%s %s %s %s [%s] \"%s\" %u %" APR_OFF_T_FMT " \"%s\" \"%s\" %s \"%s\"",
-        log_escape_nq(msr->mp, msr->hostname), msr->remote_addr, log_escape_nq(msr->mp, remote_user),
-        log_escape_nq(msr->mp, local_user), current_logtime(msr->mp),
-        ((msr->request_line == NULL) ? "" : log_escape(msr->mp, msr->request_line)),
-        msr->response_status, msr->bytes_sent, log_escape(msr->mp, referer),
-        log_escape(msr->mp, user_agent), log_escape(msr->mp, uniqueid), sessionid);
-}
-
 /**
  * Constructs a log line in vcombined log format trying to truncate
  * some of the fields to make the log line shorter than _limit bytes.
@@ -287,11 +240,17 @@ static char *create_auditlog_boundary(request_rec *r) {
  * Sanitises the request line by removing the parameters
  * that have been marked as sensitive.
  */
-static void sanitise_request_line(modsec_rec *msr) {
-    const apr_array_header_t *tarr;
-    const apr_table_entry_t *telts;
-    int i;
+static void sanitize_request_line(modsec_rec *msr) {
+    const apr_array_header_t *tarr = NULL;
+    const apr_table_entry_t *telts = NULL;
+    const apr_array_header_t *tarr_pattern = NULL;
+    const apr_table_entry_t *telts_pattern = NULL;
+    msc_parm *mparm = NULL;
+    int i, k;
     char *qspos;
+    char *buf = NULL;
+    int sanitized_partial = 0;
+    int sanitize_matched = 0;
 
     /* Locate the query string. */
     qspos = strstr(msr->request_line, "?");
@@ -299,38 +258,81 @@ static void sanitise_request_line(modsec_rec *msr) {
     qspos++;
 
     /* Loop through the list of sensitive parameters. */
-    tarr = apr_table_elts(msr->arguments_to_sanitise);
+    tarr = apr_table_elts(msr->arguments_to_sanitize);
     telts = (const apr_table_entry_t*)tarr->elts;
     for (i = 0; i < tarr->nelts; i++) {
         msc_arg *arg = (msc_arg *)telts[i].val;
         /* Only look at the parameters that appeared in the query string. */
         if (strcmp(arg->origin, "QUERY_STRING") == 0) {
+            char *pat = NULL;
             char *p;
-            int j;
+            int j, arg_min, arg_max;
 
             /* Go to the beginning of the parameter. */
             p = qspos;
             j = arg->value_origin_offset;
             while((*p != '\0')&&(j--)) p++;
             if (*p == '\0') {
-                msr_log(msr, 1, "Unable to sanitise variable \"%s\" at offset %u of QUERY_STRING"
-                    "because the request line is too short.",
-                    log_escape_ex(msr->mp, arg->name, arg->name_len),
-                    arg->value_origin_offset);
+                msr_log(msr, 1, "Unable to sanitize variable \"%s\" at offset %u of QUERY_STRING"
+                        "because the request line is too short.",
+                        log_escape_ex(msr->mp, arg->name, arg->name_len),
+                        arg->value_origin_offset);
                 continue;
             }
 
-            /* Write over the value. */
-            j = arg->value_origin_len;
-            while((*p != '\0')&&(j--)) {
-                *p++ = '*';
+            tarr_pattern = apr_table_elts(msr->pattern_to_sanitize);
+            telts_pattern = (const apr_table_entry_t*)tarr_pattern->elts;
+
+            sanitized_partial = 0;
+            sanitize_matched = 0;
+            buf = apr_psprintf(msr->mp, "%s",p);
+            for ( k = 0; k < tarr_pattern->nelts; k++)  {
+                if(strncmp(telts_pattern[k].key,arg->name,strlen(arg->name)) ==0 ) {
+                    mparm = (msc_parm *)telts_pattern[k].val;
+                    pat = strstr(buf,mparm->value);
+                    if(mparm->pad_1 == -1)
+                        sanitize_matched = 1;
+
+                    if (pat != NULL) {
+                        j = strlen(mparm->value);
+                        arg_min = j;
+                        arg_max = 1;
+                        while((*pat != '\0')&&(j--)) {
+                            if(arg_max > mparm->pad_2)  {
+                                int off = (strlen(mparm->value) - arg_max);
+                                int pos = (mparm->pad_1-1);
+                                if(off > pos)    {
+                                    *pat = '*';
+                                }
+                            }
+                            arg_max++;
+                            arg_min--;
+                            pat++;
+                        }
+                    }
+                    sanitized_partial = 1;
+                }
             }
-            if (*p == '\0') {
-                msr_log(msr, 1, "Unable to sanitise variable \"%s\" at offset %u (size %d) "
-                    "of QUERY_STRING because the request line is too short.",
-                    log_escape_ex(msr->mp, arg->name, arg->name_len),
-                    arg->value_origin_offset, arg->value_origin_len);
+
+            if(sanitized_partial == 1 && sanitize_matched == 0)  {
+                while(*buf != '\0') {
+                    *p++ = *buf++;
+                }
                 continue;
+            } else {
+                /* Write over the value. */
+                j = arg->value_origin_len;
+                while((*p != '\0')&&(j--)) {
+                    *p++ = '*';
+                }
+
+                if (*p == '\0') {
+                    msr_log(msr, 1, "Unable to sanitize variable \"%s\" at offset %u (size %d) "
+                            "of QUERY_STRING because the request line is too short.",
+                            log_escape_ex(msr->mp, arg->name, arg->name_len),
+                            arg->value_origin_offset, arg->value_origin_len);
+                    continue;
+                }
             }
         }
     }
@@ -367,21 +369,126 @@ static void sec_auditlog_write_producer_header(modsec_rec *msr) {
     sec_auditlog_write(msr, ".\n", 2);
 }
 
+/*
+* \brief This function will returns the next chain node
+*
+* \param current Pointer to current chined rule
+* \param msr Pointer to modsec_rec
+*
+* \retval NULL On failure
+* \retval next_rule On Success
+*/
+static msre_rule *return_chained_rule(const msre_rule *current, modsec_rec *msr)   {
+    apr_array_header_t *arr = NULL;
+    msre_rule **rules = NULL;
+    msre_rule *rule = NULL, *next_rule = NULL;
+    int i;
+    int phase;
+
+    if (current == NULL || current->actionset == NULL || current->ruleset == NULL)
+        return NULL;
+
+    phase = current->actionset->phase;
+
+    switch (phase) {
+        case PHASE_REQUEST_HEADERS :
+            arr = current->ruleset->phase_request_headers;
+            break;
+        case PHASE_REQUEST_BODY :
+            arr = current->ruleset->phase_request_body;
+            break;
+        case PHASE_RESPONSE_HEADERS :
+            arr = current->ruleset->phase_response_headers;
+            break;
+        case PHASE_RESPONSE_BODY :
+            arr = current->ruleset->phase_response_body;
+            break;
+        case PHASE_LOGGING :
+            arr = current->ruleset->phase_logging;
+            break;
+        default :
+            msr_log(msr, 1, "Logging: Invalid phase %d",phase);
+            return NULL;
+    }
+
+    rules = (msre_rule **)arr->elts;
+    for (i = 0; i < arr->nelts; i++) {
+        rule = rules[i];
+        if (rule != NULL)    {
+            if (strncmp(current->unparsed,rule->unparsed,strlen(current->unparsed)) == 0) {
+
+                if (i < arr->nelts -1)   {
+                    next_rule = rules[i+1];
+                } else  {
+                    next_rule = rules[i];
+                }
+
+                if (next_rule == NULL || next_rule->chain_starter == NULL)
+                    return NULL;
+
+                if(current->chain_starter == NULL && next_rule->chain_starter != NULL)  {
+                    if (strncmp(current->unparsed, next_rule->chain_starter->unparsed, strlen(current->unparsed)) != 0)
+                        return NULL;
+                }
+
+                if(current->chain_starter != NULL && next_rule->chain_starter != NULL)  {
+                    if (strncmp(current->chain_starter->unparsed, rule->chain_starter->unparsed, strlen(current->chain_starter->unparsed)) != 0)
+                        return NULL;
+                }
+
+                return next_rule;
+            }
+        }
+    }
+
+    return NULL;
+}
+
+/*
+* \brief This function will check if a chained rule
+* appears into matched array.
+*
+* \param msr Pointer to modsec_rec
+* \param next_rule Pointer to chained rule
+*
+* \retval 0 On failure
+* \retval 1 On Success
+*/
+static int chained_is_matched(modsec_rec *msr, const msre_rule *next_rule) {
+    int i = 0;
+    const msre_rule *rule = NULL;
+
+    for (i = 0; i < msr->matched_rules->nelts; i++) {
+        rule = ((msre_rule **)msr->matched_rules->elts)[i];
+        if (rule != NULL && (strncmp(rule->unparsed,next_rule->unparsed,strlen(rule->unparsed)) == 0))  {
+            return 1;
+        }
+    }
+
+    return 0;
+}
+
 /**
  * Produce an audit log entry.
  */
 void sec_audit_logger(modsec_rec *msr) {
     const apr_array_header_t *arr = NULL;
     apr_table_entry_t *te = NULL;
+    const apr_array_header_t *tarr_pattern = NULL;
+    const apr_table_entry_t *telts_pattern = NULL;
     char *str1 = NULL, *str2 = NULL, *text = NULL;
-    const msre_rule *rule = NULL;
+    const msre_rule *rule = NULL, *next_rule = NULL;
     apr_size_t nbytes, nbytes_written;
     unsigned char md5hash[APR_MD5_DIGESTSIZE];
     int was_limited = 0;
+    int present = 0;
     int wrote_response_body = 0;
     char *entry_filename, *entry_basename;
     apr_status_t rc;
-    int i, limit;
+    int i, limit, k, sanitized_partial, j;
+    char *buf = NULL, *pat = NULL;
+    msc_parm *mparm = NULL;
+    int arg_min, arg_max, sanitize_matched;
 
     /* the boundary is used by both audit log types */
     msr->new_auditlog_boundary = create_auditlog_boundary(msr->r);
@@ -447,7 +554,7 @@ void sec_audit_logger(modsec_rec *msr) {
          * the folder is there.
          */
         rc = apr_dir_make_recursive(entry_basename, msr->txcfg->auditlog_dirperms, msr->mp);
-        if (rc != APR_SUCCESS) {
+        if ((rc != APR_SUCCESS) && (rc != APR_EEXIST)) {
             msr_log(msr, 1, "Audit log: Failed to create subdirectories: %s (%s)",
                 entry_basename, get_apr_error(msr->mp, rc));
             return;
@@ -491,19 +598,56 @@ void sec_audit_logger(modsec_rec *msr) {
         text = apr_psprintf(msr->mp, "\n--%s-%c--\n", msr->new_auditlog_boundary, AUDITLOG_PART_REQUEST_HEADERS);
         sec_auditlog_write(msr, text, strlen(text));
 
-        sanitise_request_line(msr);
+        sanitize_request_line(msr);
 
         sec_auditlog_write(msr, msr->request_line, strlen(msr->request_line));
         sec_auditlog_write(msr, "\n", 1);
 
         arr = apr_table_elts(msr->request_headers);
         te = (apr_table_entry_t *)arr->elts;
+
+        tarr_pattern = apr_table_elts(msr->pattern_to_sanitize);
+        telts_pattern = (const apr_table_entry_t*)tarr_pattern->elts;
+
         for (i = 0; i < arr->nelts; i++) {
+            sanitized_partial = 0;
+            sanitize_matched = 0;
             text = apr_psprintf(msr->mp, "%s: %s\n", te[i].key, te[i].val);
-            /* Do we need to sanitise this request header? */
-            if (apr_table_get(msr->request_headers_to_sanitise, te[i].key) != NULL) {
-                /* Yes, sanitise it. */
-                memset(text + strlen(te[i].key) + 2, '*', strlen(te[i].val));
+            if (apr_table_get(msr->request_headers_to_sanitize, te[i].key) != NULL) {
+                buf = apr_psprintf(msr->mp, "%s",text+strlen(te[i].key)+2);
+
+                for ( k = 0; k < tarr_pattern->nelts; k++)  {
+                    if(strncmp(telts_pattern[k].key,te[i].key,strlen(te[i].key)) ==0 ) {
+                        mparm = (msc_parm *)telts_pattern[k].val;
+                        if(mparm->pad_1 == -1)
+                            sanitize_matched = 1;
+                        pat = strstr(buf,mparm->value);
+                        if (pat != NULL)    {
+                            j = strlen(mparm->value);
+                            arg_min = j;
+                            arg_max = 1;
+                            while((*pat != '\0')&&(j--)) {
+                                if(arg_max > mparm->pad_2)  {
+                                    int off = strlen(mparm->value) - arg_max;
+                                    int pos = mparm->pad_1-1;
+                                    if(off > pos)    {
+                                        *pat = '*';
+                                    }
+                                }
+                                arg_max++;
+                                arg_min--;
+                                pat++;
+                            }
+                            sanitized_partial = 1;
+                        }
+                    }
+                }
+
+                if(sanitized_partial == 1 && sanitize_matched == 0)  {
+                    text = apr_psprintf(msr->mp, "%s: %s\n", te[i].key, buf);
+                } else {
+                    memset(text + strlen(te[i].key) + 2, '*', strlen(te[i].val));
+                }
             }
             sec_auditlog_write(msr, text, strlen(text));
         }
@@ -525,17 +669,17 @@ void sec_audit_logger(modsec_rec *msr) {
             apr_array_header_t *sorted_args;
             unsigned int offset = 0, last_offset = 0;
             msc_arg *nextarg = NULL;
-            int sanitise = 0; /* IMP1 Use constants for "sanitise" values. */
+            int sanitize = 0; /* IMP1 Use constants for "sanitize" values. */
             char *my_error_msg = NULL;
 
             sorted_args = apr_array_make(msr->mp, 25, sizeof(const msc_arg *));
 
             /* First we need to sort the arguments that need to be
-             * sanitised in descending order (we are using a stack structure
+             * sanitized in descending order (we are using a stack structure
              * to store then so the order will be ascending when we start
              * popping them out). This is because we will
              * be reading the request body sequentially and must
-             * sanitise it as we go.
+             * sanitize it as we go.
              */
 
             for(;;) {
@@ -544,7 +688,7 @@ void sec_audit_logger(modsec_rec *msr) {
                 /* Find the next largest offset (excluding
                  * the ones we've used up already).
                  */
-                tarr = apr_table_elts(msr->arguments_to_sanitise);
+                tarr = apr_table_elts(msr->arguments_to_sanitize);
                 telts = (const apr_table_entry_t*)tarr->elts;
                 for(i = 0; i < tarr->nelts; i++) {
                     msc_arg *arg = (msc_arg *)telts[i].val;
@@ -570,7 +714,7 @@ void sec_audit_logger(modsec_rec *msr) {
                  */
                 if (nextarg == NULL) break;
 
-                sanitise = 2; /* Means time to pop the next argument out. */
+                sanitize = 2; /* Means time to pop the next argument out. */
                 last_offset = offset;
                 offset = 0;
                 { /* IMP1 Fix this ugly bit here. */
@@ -580,7 +724,7 @@ void sec_audit_logger(modsec_rec *msr) {
             }
 
             /* Now start retrieving the body chunk by chunk and
-             * sanitise data in pieces.
+             * sanitize data in pieces.
              */
 
             rc = modsecurity_request_body_retrieve_start(msr, &my_error_msg);
@@ -589,8 +733,8 @@ void sec_audit_logger(modsec_rec *msr) {
             } else {
                 msc_data_chunk *chunk = NULL;
                 unsigned int chunk_offset = 0;
-                unsigned int sanitise_offset = 0;
-                unsigned int sanitise_length = 0;
+                unsigned int sanitize_offset = 0;
+                unsigned int sanitize_length = 0;
 
                 text = apr_psprintf(msr->mp, "\n--%s-%c--\n", msr->new_auditlog_boundary, AUDITLOG_PART_REQUEST_BODY);
                 sec_auditlog_write(msr, text, strlen(text));
@@ -598,46 +742,46 @@ void sec_audit_logger(modsec_rec *msr) {
                 for(;;) {
                     rc = modsecurity_request_body_retrieve(msr, &chunk, -1, &my_error_msg);
                     if (chunk != NULL) {
-                        /* Anything greater than 1 means we have more data to sanitise. */
-                        while (sanitise > 1) {
+                        /* Anything greater than 1 means we have more data to sanitize. */
+                        while (sanitize > 1) {
                             msc_arg **arg = NULL;
 
-                            if (sanitise == 2) {
+                            if (sanitize == 2) {
                                 /* Get the next argument from the stack. */
                                 arg = (msc_arg **)apr_array_pop(sorted_args);
-                                if (arg == NULL) sanitise = 0; /* We're done sanitising. */
+                                if (arg == NULL) sanitize = 0; /* We're done sanitising. */
                                 else {
                                     /* Continue with sanitation to process the
                                      * retrieved argument.
                                      */
-                                    sanitise = 1;
-                                    sanitise_offset = (*arg)->value_origin_offset;
-                                    sanitise_length = (*arg)->value_origin_len;
+                                    sanitize = 1;
+                                    sanitize_offset = (*arg)->value_origin_offset;
+                                    sanitize_length = (*arg)->value_origin_len;
                                 }
                             }
 
-                            if (sanitise) {
-                                /* Check if the data we want to sanitise is
+                            if (sanitize) {
+                                /* Check if the data we want to sanitize is
                                  * stored in the current chunk.
                                  */
-                                if (chunk_offset + chunk->length > sanitise_offset) {
+                                if (chunk_offset + chunk->length > sanitize_offset) {
                                     unsigned int soff; /* data offset within chunk */
-                                    unsigned int len;  /* amount in this chunk to sanitise */
+                                    unsigned int len;  /* amount in this chunk to sanitize */
 
-                                    soff = sanitise_offset - chunk_offset;
+                                    soff = sanitize_offset - chunk_offset;
 
-                                    if (soff + sanitise_length <= chunk->length) {
+                                    if (soff + sanitize_length <= chunk->length) {
                                         /* The entire argument resides in the current chunk. */
-                                        len = sanitise_length;
-                                        sanitise = 2; /* Get another parameter to sanitise. */
+                                        len = sanitize_length;
+                                        sanitize = 2; /* Get another parameter to sanitize. */
                                     } else {
                                         /* Some work to do here but we'll need to seek
                                          * another chunk.
                                          */
                                         len = chunk->length - soff;
-                                        sanitise_offset += len;
-                                        sanitise_length -= len;
-                                        sanitise = 1; /* It's OK to go to the next chunk. */
+                                        sanitize_offset += len;
+                                        sanitize_length -= len;
+                                        sanitize = 1; /* It's OK to go to the next chunk. */
                                     }
 
                                     /* Yes, we actually write over the original data.
@@ -650,7 +794,7 @@ void sec_audit_logger(modsec_rec *msr) {
                             }
                         }
 
-                        /* Write the sanitised chunk to the log
+                        /* Write the sanitized chunk to the log
                          * and advance to the next chunk. */
                         sec_auditlog_write(msr, chunk->data, chunk->length);
                         chunk_offset += chunk->length;
@@ -676,7 +820,7 @@ void sec_audit_logger(modsec_rec *msr) {
         if ((msr->msc_reqbody_read)&&(msr->mpd != NULL)) {
             char *buffer = NULL;
 
-            buffer = multipart_reconstruct_urlencoded_body_sanitise(msr);
+            buffer = multipart_reconstruct_urlencoded_body_sanitize(msr);
             if (buffer == NULL) {
                 msr_log(msr, 1, "Audit log: Failed to reconstruct request body.");
             } else {
@@ -697,10 +841,10 @@ void sec_audit_logger(modsec_rec *msr) {
         if (msr->response_headers_sent) {
             if (msr->status_line != NULL) {
                 text = apr_psprintf(msr->mp, "%s %s\n", msr->response_protocol,
-                    msr->status_line);
+                        msr->status_line);
             } else {
                 text = apr_psprintf(msr->mp, "%s %u\n", msr->response_protocol,
-                    msr->response_status);
+                        msr->response_status);
             }
             sec_auditlog_write(msr, text, strlen(text));
 
@@ -708,18 +852,57 @@ void sec_audit_logger(modsec_rec *msr) {
 
             arr = apr_table_elts(msr->response_headers);
             te = (apr_table_entry_t *)arr->elts;
+
+            tarr_pattern = apr_table_elts(msr->pattern_to_sanitize);
+            telts_pattern = (const apr_table_entry_t*)tarr_pattern->elts;
+
             for (i = 0; i < arr->nelts; i++) {
+                sanitized_partial = 0;
+                sanitize_matched = 0;
                 text = apr_psprintf(msr->mp, "%s: %s\n", te[i].key, te[i].val);
-                /* Do we need to sanitise this response header? */
-                if (apr_table_get(msr->response_headers_to_sanitise, te[i].key) != NULL) {
-                    /* Yes, sanitise it. */
-                    memset(text + strlen(te[i].key) + 2, '*', strlen(te[i].val));
+                if (apr_table_get(msr->response_headers_to_sanitize, te[i].key) != NULL) {
+                    buf = apr_psprintf(msr->mp, "%s",text+strlen(te[i].key)+2);
+
+                    for ( k = 0; k < tarr_pattern->nelts; k++)  {
+                        if(strncmp(telts_pattern[k].key,te[i].key,strlen(te[i].key)) ==0 ) {
+                            mparm = (msc_parm *)telts_pattern[k].val;
+                            if(mparm->pad_1 == -1)
+                                sanitize_matched = 1;
+                            pat = strstr(buf,mparm->value);
+                            if (pat != NULL)    {
+                                j = strlen(mparm->value);
+                                arg_min = j;
+                                arg_max = 1;
+                                while((*pat != '\0')&&(j--)) {
+                                    if(arg_max > mparm->pad_2)  {
+                                        int off = strlen(mparm->value) - arg_max;
+                                        int pos = mparm->pad_1-1;
+                                        if(off > pos)    {
+                                            *pat = '*';
+                                        }
+                                    }
+                                    arg_max++;
+                                    arg_min--;
+                                    pat++;
+                                }
+                                sanitized_partial = 1;
+                            }
+                        }
+                    }
+
+                    if(sanitized_partial == 1 && sanitize_matched == 0)  {
+                        text = apr_psprintf(msr->mp, "%s: %s\n", te[i].key, buf);
+                    } else {
+                        memset(text + strlen(te[i].key) + 2, '*', strlen(te[i].val));
+                    }
                 }
                 sec_auditlog_write(msr, text, strlen(text));
             }
         }
     }
 
+    apr_table_clear(msr->pattern_to_sanitize);
+
     /* AUDITLOG_PART_RESPONSE_BODY */
 
     if (strchr(msr->txcfg->auditlog_parts, AUDITLOG_PART_RESPONSE_BODY) != NULL) {
@@ -765,36 +948,21 @@ void sec_audit_logger(modsec_rec *msr) {
             sec_auditlog_write(msr, text, strlen(text));
         }
 
-        /* Processing times */
-        if (msr->time_checkpoint_1 == 0) {
-            text = apr_psprintf(msr->mp, "Stopwatch: %" APR_TIME_T_FMT " %" APR_TIME_T_FMT
-                " (- - -)\n", (msr->request_time), (now - msr->request_time));
-        } else {
-            char sw_str2[101] = "-";
-            char sw_str3[101] = "-";
-
-            if (msr->time_checkpoint_2 != 0) {
-                apr_snprintf(sw_str2, sizeof(sw_str2), "%" APR_TIME_T_FMT,
-                    (msr->time_checkpoint_2 - msr->request_time));
-            }
-
-            if (msr->time_checkpoint_3 != 0) {
-                apr_snprintf(sw_str3, sizeof(sw_str3), "%" APR_TIME_T_FMT,
-                    (msr->time_checkpoint_3 - msr->request_time));
-            }
-
-            text = apr_psprintf(msr->mp, "Stopwatch: %" APR_TIME_T_FMT
-                " %" APR_TIME_T_FMT " (%" APR_TIME_T_FMT
-                "%s %s %s)\n",
-                (msr->request_time), (now - msr->request_time),
-                (msr->time_checkpoint_1 - msr->request_time),
-                ((msr->msc_reqbody_read == 0) ? "" : "*"),
-                sw_str2, sw_str3
-                );
-        }
-
+        /* Stopwatch; left in for compatibility reasons */
+        text = apr_psprintf(msr->mp, "Stopwatch: %" APR_TIME_T_FMT " %" APR_TIME_T_FMT " (- - -)\n",
+            msr->request_time, (now - msr->request_time));
         sec_auditlog_write(msr, text, strlen(text));
 
+        /* Stopwatch2 */
+        {
+            char *perf_all = format_all_performance_variables(msr, msr->mp);
+
+            text = apr_psprintf(msr->mp, "Stopwatch2: %" APR_TIME_T_FMT " %" APR_TIME_T_FMT
+                "; %s\n", msr->request_time, (now - msr->request_time), perf_all);
+
+            sec_auditlog_write(msr, text, strlen(text));
+        }
+
         /* Our response body does not contain chunks */
         /* ENH Only write this when the output was chunked. */
         /* ENH Add info when request body was decompressed, dechunked too. */
@@ -816,7 +984,7 @@ void sec_audit_logger(modsec_rec *msr) {
             const apr_array_header_t *tarr;
             const apr_table_entry_t *telts;
 
-            tarr = apr_table_elts(msr->arguments_to_sanitise);
+            tarr = apr_table_elts(msr->arguments_to_sanitize);
             telts = (const apr_table_entry_t*)tarr->elts;
 
             if (tarr->nelts > 0) {
@@ -837,7 +1005,7 @@ void sec_audit_logger(modsec_rec *msr) {
             const apr_array_header_t *tarr;
             const apr_table_entry_t *telts;
 
-            tarr = apr_table_elts(msr->request_headers_to_sanitise);
+            tarr = apr_table_elts(msr->request_headers_to_sanitize);
             telts = (const apr_table_entry_t*)tarr->elts;
 
             if (tarr->nelts > 0) {
@@ -857,7 +1025,7 @@ void sec_audit_logger(modsec_rec *msr) {
             const apr_array_header_t *tarr;
             const apr_table_entry_t *telts;
 
-            tarr = apr_table_elts(msr->response_headers_to_sanitise);
+            tarr = apr_table_elts(msr->response_headers_to_sanitize);
             telts = (const apr_table_entry_t*)tarr->elts;
 
             if (tarr->nelts > 0) {
@@ -885,8 +1053,27 @@ void sec_audit_logger(modsec_rec *msr) {
     }
 
     /* AUDITLOG_PART_UPLOADS */
-    /* ENH: Implement */
+    if ((strchr(msr->txcfg->auditlog_parts, AUDITLOG_PART_UPLOADS) != NULL) && (msr->mpd != NULL)) {
+        multipart_part **parts = NULL;
+        unsigned int total_size = 0;
+        int cfiles = 0;
 
+        text = apr_psprintf(msr->mp, "\n--%s-%c--\n", msr->new_auditlog_boundary, AUDITLOG_PART_UPLOADS);
+        sec_auditlog_write(msr, text, strlen(text));
+
+        parts = (multipart_part **)msr->mpd->parts->elts;
+        for(cfiles = 0; cfiles < msr->mpd->parts->nelts; cfiles++) {
+            if (parts[cfiles]->type == MULTIPART_FILE) {
+                if(parts[cfiles]->filename != NULL) {
+                    text = apr_psprintf(msr->mp, "%d,%u,\"%s\",\"%s\"\n", cfiles+1, parts[cfiles]->tmp_file_size, log_escape(msr->mp, parts[cfiles]->filename), log_escape(msr->mp, parts[cfiles]->content_type ? parts[cfiles]->content_type : "<Unknown ContentType>"));
+                    sec_auditlog_write(msr, text, strlen(text));
+                    total_size += parts[cfiles]->tmp_file_size;
+                }
+            }
+        }
+        text = apr_psprintf(msr->mp, "Total,%u\n", total_size);
+        sec_auditlog_write(msr, text, strlen(text));
+    }
 
     /* AUDITLOG_PART_MATCHEDRULES */
 
@@ -897,12 +1084,41 @@ void sec_audit_logger(modsec_rec *msr) {
         /* Matched Rules */
         for(i = 0; i < msr->matched_rules->nelts; i++) {
             rule = ((msre_rule **)msr->matched_rules->elts)[i];
-            text = apr_psprintf(msr->mp, "%s\n", rule->unparsed);
-            sec_auditlog_write(msr, text, strlen(text));
+            if ((rule != NULL) && (rule->actionset != NULL) && rule->actionset->is_chained && (rule->chain_starter == NULL)) {
+                text = apr_psprintf(msr->mp, "%s\n", rule->unparsed);
+                sec_auditlog_write(msr, text, strlen(text));
+                do {
+                    if (rule->ruleset != NULL)   {
+
+                        next_rule = return_chained_rule(rule,msr);
+
+                        if (next_rule != NULL)  {
+
+                            present = chained_is_matched(msr,next_rule);
+
+                            if (present == 0)   {
+                                text = apr_psprintf(msr->mp, "#%s\n",next_rule->unparsed);
+                            } else  {
+                                text = apr_psprintf(msr->mp, "%s\n",next_rule->unparsed);
+                                i++;
+                            }
+
+                            sec_auditlog_write(msr, text, strlen(text));
+                        }
+                    }
+                    rule = next_rule;
+                } while (rule != NULL && rule->actionset != NULL && rule->actionset->is_chained);
+                text = apr_psprintf(msr->mp, "\n");
+                sec_auditlog_write(msr, text, strlen(text));
+            } else  {
+                if ((rule != NULL) && (rule->actionset != NULL) && !rule->actionset->is_chained && (rule->chain_starter == NULL)) {
+                    text = apr_psprintf(msr->mp, "%s\n\n", rule->unparsed);
+                    sec_auditlog_write(msr, text, strlen(text));
+                }
+            }
         }
     }
 
-
     /* AUDITLOG_PART_ENDMARKER */
 
     text = apr_psprintf(msr->mp, "\n--%s-%c--\n", msr->new_auditlog_boundary, AUDITLOG_PART_ENDMARKER);
@@ -918,7 +1134,7 @@ void sec_audit_logger(modsec_rec *msr) {
         rc = apr_global_mutex_unlock(msr->modsecurity->auditlog_lock);
         if (rc != APR_SUCCESS) {
             msr_log(msr, 1, "Audit log: Failed to unlock global mutex: %s",
-                get_apr_error(msr->mp, rc));
+                    get_apr_error(msr->mp, rc));
         }
 
         return;
@@ -934,7 +1150,7 @@ void sec_audit_logger(modsec_rec *msr) {
     apr_md5_final(md5hash, &msr->new_auditlog_md5ctx);
 
     str2 = apr_psprintf(msr->mp, "%s %d %d md5:%s", msr->new_auditlog_filename, 0,
-        msr->new_auditlog_size, bytes2hex(msr->mp, md5hash, 16));
+            msr->new_auditlog_size, bytes2hex(msr->mp, md5hash, 16));
     if (str2 == NULL) return;
 
     /* We do not want the index line to be longer than 3980 bytes. */
